﻿using UnityEngine;

public enum gjDirection
{
    Up = 0,
    Right = 1,
    Down = 2,
    Left = 3,
}

[RequireComponent(typeof(Rigidbody2D))]
[RequireComponent(typeof(gjCharacterInput))]
public class gjCharacterController : MonoBehaviour
{
    // STATIC

    private static string s_HorizontalAxis = "Horizontal";
    private static string s_VeriticalAxis = "Vertical";
    private static string s_ButtonA = "ButtonA";
    private static string s_ButtonB = "ButtonB";

    private static int s_Animator_Sword = Animator.StringToHash("Sword");
    private static int s_Animator_Walking = Animator.StringToHash("Walking");
    private static int s_Animator_Direction = Animator.StringToHash("Direction");
    private static int s_Animator_Attacking = Animator.StringToHash("Attacking");

    private static string s_TrapTag = "Trap";
    private static string s_ExitTag = "Exit";

    // Serializable fields

    [Header("Movement")]

    [SerializeField]
    private float m_MoveForce = 1500;

    [Header("Equip")]

    [SerializeField]
    private bool m_Sword = false;

    [Header("Wave")]

    [SerializeField]
    private float m_WaveSourceRangeOverlap = 2.5f;
    [SerializeField]
    private float m_AngleHit = 60f;

    [Header("Animations")]

    [SerializeField]
    private Color m_DefaultColor = Color.white;
    [SerializeField]
    private Color m_ColorOnWave = Color.white;
    [SerializeField]
    private float m_WalkingSpeedThreshold = 0.1f;

    [Header("Audio")]

    [SerializeField]
    private AudioSource m_StepAudioSource = null;
    [SerializeField]
    private AudioSource m_SwordAudioSource = null;
    [SerializeField]
    private AudioSource m_SwordHitAudioSource = null;
    [SerializeField]
    private AudioSource m_CollisionAudioSource = null;
    [SerializeField]
    private AudioSource m_TrapAudioSource = null;

    [Header("Debug")]

    [SerializeField]
    private bool m_DrawGizomos = false;

    // Fields

    private LayerMask m_WaveSourcesLayerMask = 0;

    private float m_HorizontalAxis = 0f;
    private float m_VerticalAxis = 0f;

    private bool m_ButtonAPressRequested = false;
    private bool m_ButtonBPressRequested = false;

    private Vector2 m_LastDirection = Vector2.zero;
    private bool m_IsWalking = false;

    private bool m_Attacking = false;

    private int m_OriginalLayer = 0;
    private int m_LayerOnWave = 0;

    // Components

    private SpriteRenderer m_SpriteRenderer = null;
    private gjCharacterInput m_CharacterInput = null;
    private Animator m_Animator = null;
    private Rigidbody2D m_Rigidbody = null;
    private gjWaveController m_WaveController = null;

    // ACCESSORS

    public Vector2 lastDirection
    {
        get { return m_LastDirection; }
    }

    public bool sword
    {
        get
        {
            return m_Sword;
        }
    }

    // MonoBehaviour's interface

    void OnDrawGizmos()
    {
        if (!m_DrawGizomos)
            return;

        Gizmos.color = Color.red;
        Gizmos.DrawWireSphere(transform.position, m_WaveSourceRangeOverlap);

        Gizmos.color = Color.blue;
        Vector3 start = transform.position;
        Vector3 direction = new Vector3(m_LastDirection.x, m_LastDirection.y, 0f);
        Vector3 end = start + direction * 2f;
        Gizmos.DrawLine(start, end);

        Gizmos.color = Color.green;
        Vector3 direction1 = m_LastDirection.Rotate(m_AngleHit / 2f);
        Vector3 direction2 = m_LastDirection.Rotate(-m_AngleHit / 2f);
        Vector3 end1 = start + direction1 * 2f;
        Vector3 end2 = start + direction2 * 2f;
        Gizmos.DrawLine(start, end1);
        Gizmos.DrawLine(start, end2);
    }

    void Awake()
    {
        m_OriginalLayer = gameObject.layer;
        m_LayerOnWave = LayerMask.NameToLayer("Default");

        m_SpriteRenderer = GetComponent<SpriteRenderer>();
        m_CharacterInput = GetComponent<gjCharacterInput>();
        m_Animator = GetComponent<Animator>();
        m_Rigidbody = GetComponent<Rigidbody2D>();
        m_WaveController = GetComponent<gjWaveController>();

        m_WaveSourcesLayerMask = LayerMask.GetMask("WaveSources", "KnownWaveSources");

        SetSpriteRendererColor(m_DefaultColor);

        if (m_StepAudioSource != null)
        {
            m_StepAudioSource.loop = true;
            m_StepAudioSource.playOnAwake = false;
        }
    }

    void OnEnable()
    {
        Messenger.AddListener("WaveStarted", OnWaveStarted);
        Messenger.AddListener("WaveEnded", OnWaveEnded);
    }

    void OnDisable()
    {
        Messenger.RemoveListener("WaveStarted", OnWaveStarted);
        Messenger.RemoveListener("WaveEnded", OnWaveEnded);
    }

    void Update()
    {
        m_HorizontalAxis = m_CharacterInput.GetAxis(s_HorizontalAxis);
        m_VerticalAxis = m_CharacterInput.GetAxis(s_VeriticalAxis);

        m_ButtonAPressRequested = m_CharacterInput.GetButtonDown(s_ButtonA);
        m_ButtonBPressRequested = m_CharacterInput.GetButtonDown(s_ButtonB);

        UpdateAnimator();
    }

    void FixedUpdate()
    {
        // Handle movement.

        {
            float horizontalAbs = Mathf.Abs(m_HorizontalAxis);
            float verticalAbs = Mathf.Abs(m_VerticalAxis);
            if (horizontalAbs > Mathf.Epsilon || verticalAbs > Mathf.Epsilon)
            {
                Vector2 input = new Vector2(m_HorizontalAxis, m_VerticalAxis);
                m_Rigidbody.AddForce(m_MoveForce * input * Time.fixedDeltaTime);
                m_LastDirection = input.normalized;

                if (!m_IsWalking)
                {
                    m_IsWalking = true;
                    PlayStep();
                }
            }
            else
            {
                if (m_IsWalking)
                {
                    m_IsWalking = false;
                    StopStep();
                }
            }
        }

        // Handle actions.

        if (m_ButtonAPressRequested)
        {
            if (!m_Attacking && m_Sword)
            {
                PlaySword();
                m_Attacking = true;

                bool waveAlreadyRunning = (m_WaveController != null) ? m_WaveController.waveRunning : false;
                bool waveRequest = false;

                Transform waveSourceTransform = null;

                if (!waveAlreadyRunning)
                {
                    Collider2D[] colliders = Physics2D.OverlapCircleAll(transform.position, m_WaveSourceRangeOverlap, m_WaveSourcesLayerMask);

                    for (int index = 0; index < colliders.Length; ++index)
                    {
                        Collider2D collider = colliders[index];

                        if (collider == null || !collider.isTrigger)
                            continue;

                        Vector2 myPos = transform.position;
                        Vector2 statuePos = collider.transform.position;

                        Vector2 direction = statuePos - myPos;
                        direction.Normalize();

                        float angleDir = Vector2.Angle(m_LastDirection, direction);

                        float minAngleHit = -1f * (m_AngleHit / 2f);
                        float maxAngleHit = m_AngleHit / 2f;

                        if (angleDir > minAngleHit && angleDir < maxAngleHit)
                        {
                            waveRequest = true;
                            waveSourceTransform = collider.transform;
                            break;
                        }
                    }
                }

                if (waveRequest)
                {
                    PlaySwordHit();
                    if (m_WaveController != null)
                    {
                        Vector2 position = waveSourceTransform.position;
                        m_WaveController.StartWave(position);
                    }

                    Messenger.Broadcast("WaveSourceTouched", waveSourceTransform.gameObject);
                }
            }

            m_ButtonAPressRequested = false;
        }

        if (m_ButtonBPressRequested)
        {
            SetSword(!m_Sword);
            m_ButtonBPressRequested = false;
        }
    }

    void OnCollisionEnter2D(Collision2D i_Collision)
    {
        if (i_Collision.collider.tag == s_TrapTag)
        {
            PlayTrap();
            Messenger.Broadcast("Die");
        }
        else
        {
            PlayCollision();
            if (i_Collision.collider.tag == s_ExitTag)
            {
                Messenger.Broadcast("ExitReached");
            }
        }
    }
    
    // LOGIC

    public void SetSword(bool i_Sword)
    {
        m_Sword = i_Sword;
    }

    public void EndAttack()
    {
        m_Attacking = false;
    }

    // Effects

    public void PlayStep()
    {
        if (m_StepAudioSource != null)
        {
            m_StepAudioSource.Play();
        }
    }

    public void StopStep()
    {
        if (m_StepAudioSource != null)
        {
            m_StepAudioSource.Stop();
        }
    }

    public void PlaySword()
    {
        if (m_SwordAudioSource != null)
        {
            m_SwordAudioSource.Play();
        }
    }

    public void PlaySwordHit()
    {
        if (m_SwordHitAudioSource != null)
        {
            m_SwordHitAudioSource.Play();
        }
    }

    public void PlayTrap()
    {
        if (m_TrapAudioSource != null)
        {
            m_TrapAudioSource.Play();
        }
    }

    public void PlayCollision()
    {
        if (m_CollisionAudioSource != null)
        {
            m_CollisionAudioSource.Play();
        }
    }

    // INTERNALS

    private void UpdateAnimator()
    {
        if (m_Animator == null)
            return;

        m_Animator.SetBool(s_Animator_Sword, m_Sword);

        Vector2 velocity = m_Rigidbody.velocity;
        float speed = velocity.magnitude;
        float absSpeed = Mathf.Abs(speed);
        bool walking = (absSpeed > m_WalkingSpeedThreshold);
        m_Animator.SetBool(s_Animator_Walking, walking);

        gjDirection targetDirection = GetNearestDirection(lastDirection);
        int targetDirectionIndex = (int)targetDirection;
        m_Animator.SetInteger(s_Animator_Direction, targetDirectionIndex);

        m_Animator.SetBool(s_Animator_Attacking, m_Attacking);
    }

    private gjDirection GetNearestDirection(Vector2 i_Direction)
    {
        gjDirection targetDirection = gjDirection.Up;

        float dotUp = Vector2.Dot(i_Direction, Vector2.up);
        float dotRight = Vector2.Dot(i_Direction, Vector2.right);
        float dotDown = Vector2.Dot(i_Direction, Vector2.down);
        float dotLeft = Vector2.Dot(i_Direction, Vector2.left);

        if (dotRight >= dotUp && dotRight >= dotDown && dotRight >= dotLeft)
        {
            targetDirection = gjDirection.Right;
        }
        else
        {
            if (dotDown >= dotUp && dotDown >= dotRight && dotDown >= dotLeft)
            {
                targetDirection = gjDirection.Down;
            }
            else
            {
                if (dotLeft >= dotUp && dotLeft >= dotRight && dotLeft >= dotDown)
                {
                    targetDirection = gjDirection.Left;
                }
            }
        }

        return targetDirection;
    }

    // EVENTS

    private void OnWaveStarted()
    {
        gameObject.layer = m_LayerOnWave;
        SetSpriteRendererColor(m_ColorOnWave);
    }

    private void OnWaveEnded()
    {
        gameObject.layer = m_OriginalLayer;
        SetSpriteRendererColor(m_DefaultColor);
    }

    // UTILS

    private void SetSpriteRendererColor(Color i_Color)
    {
        if (m_SpriteRenderer != null)
        {
            m_SpriteRenderer.color = i_Color;
        }
    }
}
