﻿using UnityEngine;
using UnityEngine.UI;
using UnityEngine.SceneManagement;

using System.Collections.Generic;

public class gjGameManager : MonoBehaviour
{
    // Serializable fields

    [Header("Character")]

    [SerializeField]
    private GameObject m_CharacterGo = null;

    [Header("Wave Sources")]

    [SerializeField]
    private Color m_ColorOnWave = Color.white;
    [SerializeField]
    private Color m_DefaultColor = Color.white;

    [Header("Audio")]

    [SerializeField]
    private AudioClip m_GameplayClip = null;
    [SerializeField]
    private AudioSource m_GameOverAudioSource = null;
    [SerializeField]
    private AudioSource m_WinAudioSource = null;

    [Header("UI")]

    [SerializeField]
    private GameObject m_GameOver = null;
    [SerializeField]
    private GameObject m_Win = null;

    [SerializeField]
    private Button m_ProceedButton = null;
    [SerializeField]
    private Button m_RestartButton = null;
    [SerializeField]
    private Button[] m_ExitButtons = null;

    // Fields

    private int m_WaveSourcesLayer = 0;
    private int m_KnownWaveSourcesLayer = 0;

    private List<GameObject> m_KnownWaveSources = new List<GameObject>();

    // Components

    private AudioSource m_AudioPlayer = null;

    // MonoBehaviour's interface

    void Awake()
    {
        m_WaveSourcesLayer = LayerMask.NameToLayer("WaveSources");
        m_KnownWaveSourcesLayer = LayerMask.NameToLayer("KnownWaveSources");

        m_AudioPlayer = GetComponent<AudioSource>();
        if (m_AudioPlayer == null)
        {
            m_AudioPlayer = gameObject.AddComponent<AudioSource>();
        }

        m_AudioPlayer.playOnAwake = false;
        m_AudioPlayer.loop = true;
        m_AudioPlayer.clip = m_GameplayClip;

        if (m_GameOver != null)
        {
            m_GameOver.SetActive(false);
        }

        if (m_Win != null)
        {
            m_Win.SetActive(false);
        }
    }

    void OnEnable()
    {
        Messenger.AddListener("WaveStarted", OnWaveStarted);
        Messenger.AddListener("WaveEnded", OnWaveEnded);

        Messenger.AddListener<GameObject>("WaveSourceTouched", OnWaveSourceTouched);

        Messenger.AddListener("Die", OnDie);
        Messenger.AddListener("ExitReached", OnExitReached);

        if (m_ProceedButton != null)
        {
            m_ProceedButton.onClick.AddListener(OnProceedButtonClicked);
        }

        if (m_RestartButton != null)
        {
            m_RestartButton.onClick.AddListener(OnRestartButtonClicked);
        }

        for (int index = 0; index < m_ExitButtons.Length; ++index)
        {
            Button b = m_ExitButtons[index];
            if (b != null)
            {
                b.onClick.AddListener(OnExitButtonClicked);
            }
        }
    }

    void OnDisable()
    {
        Messenger.RemoveListener("WaveStarted", OnWaveStarted);
        Messenger.RemoveListener("WaveEnded", OnWaveEnded);

        Messenger.RemoveListener<GameObject>("WaveSourceTouched", OnWaveSourceTouched);

        Messenger.RemoveListener("Die", OnDie);
        Messenger.RemoveListener("ExitReached", OnExitReached);

        if (m_ProceedButton != null)
        {
            m_ProceedButton.onClick.RemoveListener(OnProceedButtonClicked);
        }

        if (m_RestartButton != null)
        {
            m_RestartButton.onClick.RemoveListener(OnRestartButtonClicked);
        }

        for (int index = 0; index < m_ExitButtons.Length; ++index)
        {
            Button b = m_ExitButtons[index];
            if (b != null)
            {
                b.onClick.RemoveListener(OnExitButtonClicked);
            }
        }
    }

    void Start()
    {
        m_AudioPlayer.Play();
    }

    // EVENTS

    private void OnWaveStarted()
    {
        for (int index = 0; index < m_KnownWaveSources.Count; ++index)
        {
            GameObject go = m_KnownWaveSources[index];

            if (go == null)
                continue;

            go.layer = m_WaveSourcesLayer;

            SpriteRenderer spriteRenderer = go.GetComponent<SpriteRenderer>();
            if (spriteRenderer != null)
            {
                spriteRenderer.color = m_ColorOnWave;
            }
        }
    }

    private void OnWaveEnded()
    {
        for (int index = 0; index < m_KnownWaveSources.Count; ++index)
        {
            GameObject go = m_KnownWaveSources[index];

            if (go == null)
                continue;

            go.layer = m_KnownWaveSourcesLayer;

            SpriteRenderer spriteRenderer = go.GetComponent<SpriteRenderer>();
            if (spriteRenderer != null)
            {
                spriteRenderer.color = m_DefaultColor;
            }
        }
    }

    private void OnWaveSourceTouched(GameObject i_WaveSourceGo)
    {
        if (i_WaveSourceGo == null)
            return;

        if (m_KnownWaveSources.Contains(i_WaveSourceGo))
            return;

        m_KnownWaveSources.Add(i_WaveSourceGo);
    }

    private void OnDie()
    {
        m_AudioPlayer.Stop();

        if (m_CharacterGo != null)
        {
            gjCharacterInput characterInput = m_CharacterGo.GetComponent<gjCharacterInput>();
            if (characterInput != null)
            {
                characterInput.SetAllowed(false);
            }
        }

        ShowGameOver();
    }

    private void OnExitReached()
    {
        m_AudioPlayer.Stop();

        if (m_CharacterGo != null)
        {
            gjCharacterInput characterInput = m_CharacterGo.GetComponent<gjCharacterInput>();
            if (characterInput != null)
            {
                characterInput.SetAllowed(false);
            }
        }

        ShowWin();
    }

    // UTILS

    private void ShowGameOver()
    {
        if (m_GameOver != null)
        {
            m_GameOver.SetActive(true);
        }

        if (m_GameOverAudioSource != null)
        {
            m_GameOverAudioSource.Play();
        }
    }

    private void ShowWin()
    {
        if (m_Win != null)
        {
            m_Win.SetActive(true);
        }

        if (m_WinAudioSource != null)
        {
            m_WinAudioSource.Play();
        }
    }

    // UI

    private void OnExitButtonClicked()
    {
        SceneManager.LoadScene("Menu", LoadSceneMode.Single);
    }

    private void OnProceedButtonClicked()
    {
        SceneManager.LoadScene("Menu", LoadSceneMode.Single);
    }

    private void OnRestartButtonClicked()
    {
        SceneManager.LoadScene("Game", LoadSceneMode.Single);
    }
}