﻿using UnityEngine;
using UnityEngine.UI;
using UnityEngine.SceneManagement;

using System;
using System.Collections;

[Serializable]
public class gjClipSequenceInfo
{
    [SerializeField]
    private Sprite m_SpriteLeft = null;
    [SerializeField]
    private Sprite m_SpriteRight = null;
    [SerializeField]
    private float m_Duration = 2.5f;

    public Sprite spriteLeft
    {
        get
        {
            return m_SpriteLeft;
        }
    }

    public Sprite spriteRight
    {
        get
        {
            return m_SpriteRight;
        }
    }

    public float duration
    {
        get
        {
            return m_Duration;
        }
    }
}

public class gjSequenceManager : MonoBehaviour
{
    // STATIC

    private static string s_TutorialScene = "Tutorial";

    // Serializable fields

    [SerializeField]
    private Image m_Background = null;
    [SerializeField]
    private Image m_ImageLeft = null;
    [SerializeField]
    private Image m_ImageRight = null;
    [SerializeField]
    private gjClipSequenceInfo[] m_Clips = null;

    [SerializeField]
    private float m_FadeInTime = 0f;
    [SerializeField]
    private float m_FadeOutTime = 0f;

    [SerializeField]
    private float m_TimeBetweenSprites = 0.5f;

    // ACCESSORS

    public int clipCount
    {
        get
        {
            if (m_Clips != null)
            {
                return m_Clips.Length;
            }

            return 0;
        }
    }

    // MonoBehaviour's interface

    void Start()
    {
        StartCoroutine(Sequence(OnSequenceCompleted));
    }

    // INTERNALS

    private IEnumerator Sequence(Action i_CompletedCallback = null)
    {
        for (int index = 0; index < clipCount; ++index)
        {
            gjClipSequenceInfo clipInfo = m_Clips[index];

            if (clipInfo == null)
                continue;

            SetImageSprite(m_ImageLeft, clipInfo.spriteLeft);
            SetImageSprite(m_ImageRight, clipInfo.spriteRight);

            // Fade in.

            {
                float duration = m_FadeInTime;
                float time = 0f;

                while (time < duration)
                {
                    float t = Mathf.Clamp01(time / duration);
                    SetImageAlpha(m_ImageLeft, t);
                    SetImageAlpha(m_ImageRight, t);

                    yield return null;
                    time += Time.deltaTime;
                }
            }

            // Rest

            {
                float duration = clipInfo.duration;
                yield return new WaitForSeconds(duration);
            }

            // Fade out.

            {
                float duration = m_FadeOutTime;
                float time = 0f;

                while (time < duration)
                {
                    float t = Mathf.Clamp01(time / duration);
                    SetImageAlpha(m_ImageLeft, 1f - t);
                    SetImageAlpha(m_ImageRight, 1f - t);

                    if (index + 1 >= clipCount)
                    {
                        SetImageAlpha(m_Background, 1f - t);

                    }

                    yield return null;
                    time += Time.deltaTime;
                }
            }

            // Clear.

            SetImageSprite(m_ImageLeft, null);
            SetImageAlpha(m_ImageLeft, 0f);

            SetImageSprite(m_ImageRight, null);
            SetImageAlpha(m_ImageRight, 0f);

            yield return new WaitForSeconds(m_TimeBetweenSprites);
        }

        if (i_CompletedCallback != null)
        {
            i_CompletedCallback();
        }
    }

    private void OnSequenceCompleted()
    {
        LoadNextScene();
    }

    private void LoadNextScene()
    {
        SceneManager.LoadScene(s_TutorialScene, LoadSceneMode.Single);
    }

    // UTILS

    private void SetImageSprite(Image i_Image, Sprite i_Sprite)
    {
        if (i_Image != null)
        {
            i_Image.sprite = i_Sprite;
        }
    }

    private void SetImageAlpha(Image i_Image, float i_Alpha)
    {
        if (i_Image != null)
        {
            Color currentColor = i_Image.color;
            currentColor.a = i_Alpha;
            i_Image.color = currentColor;
        }
    }
}